#define _LARGEFILE64_SOURCE
#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <regex.h>
#include <fcntl.h>

#include "array.h"
#include "misc.h"
#include "br.h"
extern "C" {
#include "mem.h"
}

#define max(x, y)	((x) > (y) ? (x) : (y))

#define M_FIELD	0	/* a selected field */
#define M_TOTAL	1	/* use whole line */
#define M_RANGE 2	/* range of characters */
#define M_REGEXP 3	/* string matched by regular expression */

void show_version(void)
{
	printf("genstats v" VERSION ", (C) 2004-2005 by folkert@vanheusden.com\n");
}

void usage(void)
{
	show_version();

	printf("\n");

	printf("-i file       read from 'file' instead of stdin\n");
	printf("-c ...        select the characters-range to look at, see below\n");
	printf("-f field      which field to look at\n");
	printf("-d delimiter  field separator\n");
	printf("-e regexp     select the part of a string which matches to the reg.exp.\n");
	printf("-E regexp     only evaluate the lines matching to the regular expression\n");
	printf("-S            simple output: field and count only\n");
	printf("-s            do not counter fields without a selected field or\n");
	printf("              without a selected range\n");
	printf("-v            verbose output\n");
	printf("-I            be case insensitive\n");
	printf("-h            this help\n");
	printf("-V            show the version and exit\n");
	printf("\n");
	printf("The range (for -c) can be:\n");
	printf("N\n");
	printf("N-\n");
	printf("N-M\n");
	printf("-M\n");
	printf("Warning: NOT including character at index M!\n");
	printf("Ranges start at offset 0 (being the first character of a line).\n");
}

int main(int argc, char *argv[])
{
	array counter(3);
	int fd = 0;
	char delimiter = ' ';
	int field_nr = 0;
	char count_always = 1;
	int range_start = -1, range_end = -1;
	char verbose = 0;
	char case_insensitive = 0;
	int total_n_lines = 0;
	int c;
	int mode = M_TOTAL;
	char *line;
	int loop;
	regex_t re_select_line, re_select_string;
	char use_re_select_line = 0;
	char simple = 0;

	/* parse commandline parameters */
	while((c = getopt(argc, argv, "Si:f:d:sc:e:E:vIhV")) != -1)
	{
		switch(c)
		{
		case 'S':
			simple = 1;
			break;

		case 'i':
			fd = open64(optarg, O_RDONLY);
			if (fd == -1)
			{
				fprintf(stderr, "Failed to open %s. Reason: %s\n", optarg, strerror(errno));
				return 1;
			}
			break;

		case 'f':
			if (mode == M_RANGE)
			{
				fprintf(stderr, "-f is an invalid option when selecting ranges.\n");
				return 1;
			}
			mode = M_FIELD;
			field_nr = atoi(optarg);
			if (field_nr < 0)
			{
				fprintf(stderr, "The field nummer cannot be a negative (%d) value.\n", field_nr);
				return 1;
			}
			else if (field_nr == 0)
			{
				fprintf(stderr, "Field 0 does not exist. The first field is '1'.\n");
				return 1;
			}
			break;

		case 'd':
			if (mode == M_RANGE)
			{
				fprintf(stderr, "-d is an invalid option when selecting ranges.\n");
				return 1;
			}
			mode = M_FIELD;
			delimiter = optarg[0];
			break;

		case 's':
			count_always = 0;
			break;

		case 'c':
			if (mode == M_FIELD)
			{
				fprintf(stderr, "-c is an invalid option when selecting fields.\n");
				return 1;
			}
			mode = M_RANGE;

			if (optarg[0] == '-')
			{
				range_start = 0;
				range_end = atoi(&optarg[1]);
			}
			else
			{
				char *min = strchr(optarg, '-');
				if (!min)
				{
					range_start = atoi(optarg);
					range_end = range_start + 1;
				}
				else
				{
					range_start = atoi(optarg);

					if (min == &optarg[strlen(optarg) - 1])
					{
						range_end = -1;
					}
					else
					{
						range_end = atoi(min + 1);
					}
				}
			}
			break;

		case 'e':
			if (mode != M_TOTAL)
			{
				fprintf(stderr, "Cannot use stringmatching together with other selection methods.\n");
				return 1;
			}
			if (regcomp(&re_select_string, optarg, REG_EXTENDED | (case_insensitive ? REG_ICASE : 0)))
			{
				fprintf(stderr, "Compilation of (extended) regular expression '%s' failed (string selection).\n", optarg);
				return 1;
			}
			mode = M_REGEXP;
			break;

		case 'E':
			if (regcomp(&re_select_line, optarg, REG_EXTENDED | (case_insensitive ? REG_ICASE : 0)))
			{
				fprintf(stderr, "Compilation of (extended) regular expression '%s' failed (line selection).\n", optarg);
				return 1;
			}
			use_re_select_line = 1;
			break;

		case 'v':
			verbose = 1;
			break;

		case 'I':
			case_insensitive = 1;
			break;

		case 'V':
			show_version();
			return 1;

		case '?':
		case 'h':
			usage();
			return 1;
		}
	}

	if (verbose)
	{
		if (fd == 0)
		{
			fprintf(stderr, "Reading from STDIN.\n");
		}

		if (mode == M_RANGE)
		{
			fprintf(stderr, "Start: %d, end: %d\n", range_start, range_end);
		}
		else if (mode == M_FIELD)
		{
			fprintf(stderr, "Field number: %d, delimiter: '%c'\n", field_nr, delimiter);
		}
	}

	/* create object for easy reading of \n terminated lines */
	buffered_reader br(fd);

	/* main loop */
	while((line = br.read_line()))
	{
		char *cur_string = line;

		/* are we only looking at lines that match some regular expression? */
		if (use_re_select_line && regexec(&re_select_line, cur_string, 0, NULL, 0) == REG_NOMATCH)
		{
			/* yes we do, and the regexp currently doesn't match */
			continue;
		}

		total_n_lines++;

		if (mode == M_FIELD)
		{
			int cur_field_nr = field_nr - 1;

			/* search selected field */
			while(cur_field_nr > 0 && cur_string != NULL)
			{
				cur_string = strchr(cur_string, delimiter);

				if (likely(cur_string))
				{
					while(*cur_string == delimiter)
						cur_string++;
				}

				cur_field_nr--;
			}

			/* field not found? (short line) */
			if (!cur_string)
			{
				if (count_always)
					cur_string = "";	/* use empty string */
				else
					cur_string = NULL;
			}
			else
			{
				/* field found, end field with 0x00 */
				char *end = strchr(cur_string, delimiter);
				if (end)
					*end = 0x00;
			}
		}
		else if (mode == M_TOTAL)
		{
			/* we already know the pointer */
		}
		else if (mode == M_RANGE)
		{
			int str_len = strlen(cur_string);

			if (count_always == 0 && max(range_start, range_end) > str_len)
			{
				cur_string = NULL;
			}
			else
			{
				if (range_end != -1)
				{
					if (range_end < str_len)
						cur_string[range_end] = 0x00;
				}

				if (range_start > str_len)
					cur_string = "";
				else
					cur_string += range_start;
			}
		}
		else if (mode == M_REGEXP)
		{
			regmatch_t smatches[1];

			if (regexec(&re_select_string, cur_string, 1, smatches, 0) == 0)
			{
				if (smatches[0].rm_so == -1)
				{
					fprintf(stderr, "Internal error in regular expression library: returned 'found something' while returning no valid offsets.\n");
					return 1;
				}

				/* terminate string */
				cur_string[smatches[0].rm_eo] = 0x00;
				/* move offset to start of matching part */
				cur_string += smatches[0].rm_so;
			}
			else
			{
				/* no match found for regexp. use whole line of not switched
				 * of with -s
				 */
				if (count_always == 0)
					cur_string = "";
			}
		}

		/* add string */
		if (cur_string)
		{
			int index = counter.addstring(cur_string, 1, case_insensitive);
			int prev_line_nr = counter.getcounter(index, 1);
			counter.addcounter(index, 2, total_n_lines - prev_line_nr);
			counter.setcounter(index, 1, total_n_lines);
		}

		free(line);
	}

	/* sort */
	counter.sort(0);

	/* show stats */
	if (verbose)
	{
		int ncols;
		char *dummy = getenv("COLUMNS");

		if (!dummy)
			ncols = 79;
		else
			ncols = atoi(dummy);

		printf("Total number of lines: %d\n\n", total_n_lines);
		printf("Index Count Percentage Interval String\n");
		for(loop=0; loop<ncols; loop++)
			printf("-");
		printf("\n");
	}

	for(loop=0; loop<counter.getn(); loop++)
	{
		if (simple)
		{
			printf("%5ld %s\n", counter.getcounter(loop, 0), counter.getstring(loop));
		}
		else
		{
			printf("%5d %5ld %9.2f%% %8.2f %s\n",
				loop + 1,
				counter.getcounter(loop, 0),
				double(counter.getcounter(loop, 0) * 100) / double(total_n_lines),
				double(counter.getcounter(loop, 2)) / double(counter.getcounter(loop, 0)),
			counter.getstring(loop));
		}
	}

	return 0;
}
